@@ -257,10 +257,6 @@ const ChatView = new Lang.Class({
         this.connect('state-flags-changed',
                      Lang.bind(this, this._updateToplevel));
         this.connect('scroll-event', Lang.bind(this, this._onScroll));
-        this.connect('edge-reached', (w, pos) => {
-            if (pos == Gtk.PositionType.BOTTOM)
-                this._autoscroll = true;
-        });
 
         this.vadjustment.connect('value-changed',
                                  Lang.bind(this, this._onValueChanged));
@@ -311,7 +307,8 @@ const ChatView = new Lang.Class({
         this._logWalker.get_events_async(NUM_INITIAL_LOG_EVENTS,
                                          Lang.bind(this, this._onLogEventsReady));
 
-        this._autoscroll = true;
+        let adj = this.vadjustment;
+        this._scrollBottom = adj.upper - adj.page_size;
 
         this._app = Gio.Application.get_default();
         PasteManager.DropTargetIface.addTargets(this, this._view);
@@ -379,6 +376,19 @@ const ChatView = new Lang.Class({
         });
     },
 
+    _foreachNickTag: function(func) {
+        let tagTable = this._view.get_buffer().get_tag_table();
+        tagTable.foreach(function(tag) {
+            if (tag._contacts)
+                func(tag);
+        });
+    },
+
+    _resetNickTag: function(nickTag) {
+        nickTag._contacts = [];
+        this._updateTagStatus(nickTag);
+    },
+
     _onStyleUpdated: function() {
         let context = this.get_style_context();
         context.save();
@@ -433,11 +443,7 @@ const ChatView = new Lang.Class({
             }
         });
 
-        let offset = NICKTAG_PREFIX.length;
-        tagTable.foreach(Lang.bind(this, function(tag) {
-            if (tag._status)
-                this._setNickStatus(tag.name.substring(offset), tag._status);
-        }));
+        this._foreachNickTag(t => { this._updateTagStatus(t); });
     },
 
     vfunc_destroy: function() {
@@ -492,7 +498,6 @@ const ChatView = new Lang.Class({
                             messageType: pending[i].get_message_type(),
                             shouldHighlight: false };
             this._insertMessage(iter, message, state);
-            this._setNickStatus(message.nick, Tp.ConnectionPresenceType.OFFLINE);
 
             if (!iter.is_end() || i < pending.length - 1)
                 this._view.buffer.insert(iter, '\n', -1);
@@ -504,13 +509,10 @@ const ChatView = new Lang.Class({
         if (this._room.type == Tp.HandleType.ROOM) {
             let members = this._channel.group_dup_members_contacts();
             for (let j = 0; j < members.length; j++)
-                this._setNickStatus(members[j].get_alias(),
-                                    Tp.ConnectionPresenceType.AVAILABLE);
+                this._trackContact(members[j]);
         } else {
-                this._setNickStatus(this._channel.connection.self_contact.get_alias(),
-                                    Tp.ConnectionPresenceType.AVAILABLE);
-                this._setNickStatus(this._channel.target_contact.get_alias(),
-                                    Tp.ConnectionPresenceType.AVAILABLE);
+                this._trackContact(this._channel.connection.self_contact);
+                this._trackContact(this._channel.target_contact);
         }
     },
 
@@ -569,19 +571,19 @@ const ChatView = new Lang.Class({
     },
 
     _updateScroll: function() {
-        if (!this._autoscroll)
-            return;
-
-        if (this._nPending == 0) {
-            this._view.emit('move-cursor',
-                            Gtk.MovementStep.BUFFER_ENDS, 1, false);
-        } else {
-            let id = Object.keys(this._pending).sort(function(a, b) {
-                return a - b;
-            })[0];
-            this._autoscroll = false;
-            this._view.scroll_mark_onscreen(this._pending[id]);
+        let adj = this.vadjustment;
+        if (adj.value == this._scrollBottom) {
+            if (this._nPending == 0) {
+                this._view.emit('move-cursor',
+                                Gtk.MovementStep.BUFFER_ENDS, 1, false);
+            } else {
+                let id = Object.keys(this._pending).sort(function(a, b) {
+                    return a - b;
+                })[0];
+                this._view.scroll_mark_onscreen(this._pending[id]);
+            }
         }
+        this._scrollBottom = adj.upper - adj.page_size;
     },
 
     _onScroll: function(w, event) {
@@ -593,21 +595,32 @@ const ChatView = new Lang.Class({
         if (hasDeltas && dy >= 0)
             return Gdk.EVENT_PROPAGATE;
 
-        this._autoscroll = false;
-
         return this._fetchBacklog();
     },
 
     _onKeyPress: function(w, event) {
         let [, keyval] = event.get_keyval();
+
+        if (keyval === Gdk.KEY_Home ||
+            keyval === Gdk.KEY_KP_Home) {
+            this._view.emit('move-cursor',
+                            Gtk.MovementStep.BUFFER_ENDS,
+                            -1, false);
+            return Gdk.EVENT_STOP;
+        } else if (keyval === Gdk.KEY_End ||
+                   keyval === Gdk.KEY_KP_End) {
+            this._view.emit('move-cursor',
+                            Gtk.MovementStep.BUFFER_ENDS,
+                            1, false);
+            return Gdk.EVENT_STOP;
+        }
+
         if (keyval != Gdk.KEY_Up &&
             keyval != Gdk.KEY_KP_Up &&
             keyval != Gdk.KEY_Page_Up &&
             keyval != Gdk.KEY_KP_Page_Up)
             return Gdk.EVENT_PROPAGATE;
 
-        this._autoscroll = false;
-
         return this._fetchBacklog();
     },
 
@@ -644,13 +657,8 @@ const ChatView = new Lang.Class({
 
     _pendingMessageRemoved: function(channel, message) {
         let [id,] = message.get_pending_message_id();
-        let mark = this._pending[id];
-        if (!mark)
-            return;
-        // Re-enable auto-scrolling if this is the most recent message
-        if (this._view.buffer.get_iter_at_mark(mark).is_end())
-            this._autoscroll = true;
-        this._view.buffer.delete_mark(mark);
+        if (this._pending[id])
+            this._view.buffer.delete_mark(this._pending[id]);
         this._app.withdraw_notification('pending-message-' + id);
         delete this._pending[id];
     },
@@ -769,17 +777,38 @@ const ChatView = new Lang.Class({
         return NICKTAG_PREFIX + Polari.util_get_basenick(nick);
     },
 
-    _setNickStatus: function(nick, status) {
-        let nickTag = this._lookupTag(this._getNickTagName(nick));
+    _trackContact: function(contact) {
+        let nickTag = this._lookupTag(this._getNickTagName(contact.alias));
         if (!nickTag)
            return;
 
-        if (status == Tp.ConnectionPresenceType.AVAILABLE)
-           nickTag.foreground_rgba = this._activeNickColor;
-        else
-           nickTag.foreground_rgba = this._inactiveNickColor;
+        let alreadyTracked = nickTag._contacts.some(c => c.alias == contact.alias);
 
-        nickTag._status = status;
+        if (!alreadyTracked)
+            nickTag._contacts.push(contact);
+
+        this._updateTagStatus(nickTag);
+    },
+
+    _untrackContact: function(contact) {
+        let nickTag = this._lookupTag(this._getNickTagName(contact.alias));
+        if (!nickTag)
+            return;
+
+        let indexToDelete = nickTag._contacts.map(c => c.alias).indexOf(contact.alias);
+
+        if (indexToDelete > -1) {
+            nickTag._contacts.splice(indexToDelete, 1);
+
+            this._updateTagStatus(nickTag);
+        }
+    },
+
+    _updateTagStatus: function(tag) {
+        if (tag._contacts.length == 0)
+            tag.foreground_rgba = this._inactiveNickColor;
+        else
+            tag.foreground_rgba = this._activeNickColor;
     },
 
     _onChannelChanged: function() {
@@ -798,6 +827,19 @@ const ChatView = new Lang.Class({
                                  : this._room.account.nickname;
         this._updateMaxNickChars(nick.length);
 
+        if (this._channel) {
+            if (this._room.type == Tp.HandleType.ROOM) {
+                let members = this._channel.group_dup_members_contacts();
+                for (let j = 0; j < members.length; j++)
+                    this._trackContact(members[j]);
+            } else {
+                this._trackContact(this._channel.connection.self_contact);
+                this._trackContact(this._channel.target_contact);
+            }
+        } else {
+            this._foreachNickTag(t => { this._resetNickTag(t); });
+        }
+
         if (!this._channel)
             return;
 
@@ -820,25 +862,13 @@ const ChatView = new Lang.Class({
                 this._insertTpMessage(this._room, message);
             }));
         this._checkMessages();
-
-        if (this._room.type == Tp.HandleType.ROOM) {
-            let members = this._channel.group_dup_members_contacts();
-            for (let j = 0; j < members.length; j++)
-                this._setNickStatus(members[j].get_alias(),
-                                    Tp.ConnectionPresenceType.AVAILABLE);
-        } else {
-                this._setNickStatus(this._channel.connection.self_contact.get_alias(),
-                                    Tp.ConnectionPresenceType.AVAILABLE);
-                this._setNickStatus(this._channel.target_contact.get_alias(),
-                                    Tp.ConnectionPresenceType.AVAILABLE);
-        }
     },
 
     _onMemberRenamed: function(room, oldMember, newMember) {
         let text = _("%s is now known as %s").format(oldMember.alias, newMember.alias);
         this._insertStatus(text, oldMember.alias, 'renamed');
-        this._setNickStatus(oldMember.alias, Tp.ConnectionPresenceType.OFFLINE);
-        this._setNickStatus(newMember.alias, Tp.ConnectionPresenceType.AVAILABLE);
+        this._untrackContact(oldMember);
+        this._trackContact(newMember);
     },
 
     _onMemberDisconnected: function(room, member, message) {
@@ -846,7 +876,7 @@ const ChatView = new Lang.Class({
         if (message)
             text += ' (%s)'.format(message);
         this._insertStatus(text, member.alias, 'left');
-        this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
+        this._untrackContact(member);
     },
 
     _onMemberKicked: function(room, member, actor) {
@@ -855,7 +885,7 @@ const ChatView = new Lang.Class({
                                                          actor.alias)
                   : _("%s has been kicked").format(member.alias);
         this._insertStatus(message, member.alias, 'left');
-        this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
+        this._untrackContact(member);
     },
 
     _onMemberBanned: function(room, member, actor) {
@@ -864,13 +894,13 @@ const ChatView = new Lang.Class({
                                                          actor.alias)
                   : _("%s has been banned").format(member.alias)
         this._insertStatus(message, member.alias, 'left');
-        this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
+        this._untrackContact(member);
     },
 
     _onMemberJoined: function(room, member) {
         let text = _("%s joined").format(member.alias);
         this._insertStatus(text, member.alias, 'joined');
-        this._setNickStatus(member.alias, Tp.ConnectionPresenceType.AVAILABLE);
+        this._trackContact(member);
     },
 
     _onMemberLeft: function(room, member, message) {
@@ -880,7 +910,7 @@ const ChatView = new Lang.Class({
             text += ' (%s)'.format(message);
 
         this._insertStatus(text, member.alias, 'left');
-        this._setNickStatus(member.alias, Tp.ConnectionPresenceType.OFFLINE);
+        this._untrackContact(member);
     },
 
     _onMessageReceived: function(room, tpMessage) {
@@ -1114,7 +1144,7 @@ const ChatView = new Lang.Class({
 
         let iter = this._view.buffer.get_end_iter();
         this._insertMessage(iter, message, this._state);
-        this._setNickStatus(message.nick, Tp.ConnectionPresenceType.AVAILABLE);
+        this._trackContact(tpMessage.sender);
 
         let [id, valid] = tpMessage.get_pending_message_id();
 
@@ -1200,6 +1230,8 @@ const ChatView = new Lang.Class({
                 if (!nickTag) {
                     nickTag = new Gtk.TextTag({ name: nickTagName });
                     this._view.get_buffer().get_tag_table().add(nickTag);
+
+                    this._resetNickTag(nickTag);
                 }
                 tags.push(nickTag);
                 if (needsGap)
